/*

The example code {in this document / on this webpage} (hereafter Example 
Code) is subject to all of the following terms and conditions:

1. Intel Corporation ("Intel") grants to you a non-exclusive, non-
assignable copyright license to make only the minimum number of copies of 
the Example Code reasonably necessary for your internal use, and to 
modify the Example Code that are provided in source code (human readable) 
form.

2. You may not reverse-assemble, reverse-compile, or otherwise reverse-
engineer any software provided solely in binary form.
 
3. You may not distribute to any third party any portion of the Example 
Code in any form.
 
4. Title to the Example Code and all copies thereof remain with Intel or 
its suppliers.  The Example Code are copyrighted and are protected by 
United States copyright laws and international treaty provisions.  You 
will not remove any copyright notice from the Example Code.  You agree to 
prevent any unauthorized copying of the Example Code.  Except as 
expressly provided herein, Intel does not grant any express or implied 
right to you under Intel patents, copyrights, trademarks, or trade secret 
information.  Subject to Intels ownership of the Example Code, all 
right, title and interest in and to your modifications shall belong to 
you.
 
5. The Example Code are provided "AS IS" without warranty of any kind.  
INTEL OFFERS NO WARRANTY EITHER EXPRESS OR IMPLIED INCLUDING THOSE OF 
MERCHANTABILITY, NONINFRINGEMENT OF THIRD- PARTY INTELLECTUAL PROPERTY OR 
FITNESS FOR A PARTICULAR PURPOSE. NEITHER INTEL NOR ITS SUPPLIERS SHALL 
BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING, WITHOUT LIMITATION, 
DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF 
BUSINESS INFORMATION, OR OTHER LOSS) ARISING OUT OF THE USE OF OR 
INABILITY TO USE THE EXAMPLE CODE, EVEN IF INTEL HAS BEEN ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGES.  BECAUSE SOME JURISDICTIONS PROHIBIT THE 
EXCLUSION OR LIMITATION OF LIABILITY FOR CONSEQUENTIAL OR INCIDENTAL 
DAMAGES, THE ABOVE LIMITATION MAY NOT APPLY TO YOU.
 
6. THE EXAMPLE CODE ARE NOT DESIGNED, INTENDED, OR AUTHORIZED FOR USE IN 
ANY TYPE OF SYSTEM OR APPLICATION IN WHICH THE FAILURE OF THE EXAMPLE 
CODE COULD CREATE A SITUATION WHERE PERSONAL INJURY OR DEATH MAY OCCUR 
(E.G MEDICAL SYSTEMS, LIFE SUSTAINING OR LIFE SAVING SYSTEMS).  Should 
you use the Example Code for any such unintended or unauthorized use, you 
shall indemnify and hold Intel and its officers, subsidiaries and 
affiliates harmless against all claims, costs, damages, and expenses, and 
reasonable attorney fees arising out of, directly or indirectly, any 
claim of product liability, personal injury or death associated with such 
unintended or unauthorized use, even if such claim alleges that Intel was 
negligent regarding the design or manufacture of the part.
 
7. You agree that any material, information or other communication you 
transmit or post to an Intel website or provide to Intel under this 
Agreement will be considered non-confidential and non-proprietary 
("Communications").  Intel will have no obligations with respect to the 
Communications.  You agree that Intel and its designees will be free to 
copy, modify, create derivative works, publicly display, disclose, 
distribute, license and sublicense through multiple tiers of distribution 
and licensees, incorporate and otherwise use the Communications and all 
data, images, sounds, text, and other things embodied therein, including 
derivative works thereto, for any and all commercial or non-commercial 
purposes. You are prohibited from posting or transmitting to or from an 
Intel website or provide to Intel any unlawful, threatening, libelous, 
defamatory, obscene, pornographic, or other material that would violate 
any law.  If you wish to provide Intel with your confidential 
information, Intel requires a non-disclosure agreement (NDA) to receive 
such confidential information, so please contact your Intel 
representative to ensure the proper NDA is in place.
 
8. The term of this Agreement will commence on the date this Agreement is 
accepted by you and will continue until terminated.  You may terminate 
this Agreement at any time.  Intel may terminate this Agreement at any 
time with written notice to you should you breach any provision in this 
Agreement.  Upon termination, you will immediately destroy the Example 
Code along with any copies you have made.
 
9. U.S. GOVERNMENT RESTRICTED RIGHTS: The Example Code are provided with 
"RESTRICTED RIGHTS".  Use, duplication or disclosure by the Government is 
subject to restrictions set forth in FAR52.227-14 and DFAR252.227-7013 et 
seq. or its successor.  Use of the Materials by the Government 
constitutes acknowledgment of Intel's rights in them.
 
10. APPLICABLE LAWS: Any claim arising under or relating to this 
Agreement shall be governed by the internal substantive laws of the State 
of Delaware or federal courts located in Delaware, without regard to 
principles of conflict of laws.  You may not export the Example Code in 
violation of applicable export laws.

* Other names and brands may be claimed as the property of others

*/

#include "parameter.h"
#include "grid.h"
#include <malloc.h>
#include <stdio.h>
#include <math.h>

#ifdef USE_MPI

/*
 *
 *  Communication related routines
 *
 */


float* left_col;
float* right_col;


void process_grid(para* p, grid* gr)
{

  float eps = 1.e-5f;
  int proc_rows, proc_columns;

  int rows, columns;
  int start_row, start_column;

  int nprocs = p->nprocs;
  int me     = p->me;

  int my_row=0, my_column=0;

  int up_rank,down_rank,right_rank,left_rank;

  // compute position in process grid
  // first compute number of process rows in 2D grid
  // should be equal or larger than process columns

  proc_columns = (int) (sqrt((float) nprocs) + eps);

  //printf("proc_columns = %d\n", proc_columns);

  while( proc_columns != 1 )
  {
     if( (nprocs % proc_columns) == 0) break;
     proc_columns--;
  }

  proc_rows = nprocs / proc_columns;

  // if processor rows are set change defaults distribution

  if(p->prows)
  {
     int col = nprocs/ p->prows;

     if( nprocs == col * p->prows)
     {
       proc_rows    = p->prows;
       proc_columns = col;
     }

  }

  if(me==0)
    printf("number of procs: %d = %d x %d\n", nprocs, proc_rows, proc_columns);

  my_row    = me/proc_columns;
  my_column = me % proc_columns;

  //printf("my row,column = %d,%d\n", my_row,my_column);

  // rank of upper process

  if( my_row == (proc_rows-1))
    up_rank = MPI_PROC_NULL;
  else
    up_rank = (my_row+1)*proc_columns + my_column;

  // rank of lower process;

  if( my_row == 0)
    down_rank = MPI_PROC_NULL;
  else
    down_rank = (my_row-1)*proc_columns + my_column;

 // rank of right process

  if( my_column == (proc_columns-1))
    right_rank = MPI_PROC_NULL;
  else
    right_rank = me + 1;

  // rank of left process;

  if( my_column == 0)
    left_rank = MPI_PROC_NULL;
  else
    left_rank = me -1;

  //printf("up, down, left, right process = %d,%d,%d,%d \n",up_rank, down_rank,
  //                                                      left_rank, right_rank);

  gr->up    = up_rank;
  gr->down  = down_rank;
  gr->left  = left_rank;
  gr->right = right_rank;

 /*

       number of rows  and columns
       start row index
       start column index

*/

  {
    int n        = p->n;
    int m        = p->m;

    int num_rows, num_columns;
    int i_start, j_start;

    num_rows = n/proc_rows;

    if( my_row < n%proc_rows)
    {
          num_rows++;
          i_start = num_rows * my_row + 1;
    }
    else
         i_start  = num_rows*my_row + n%proc_rows +1;

    num_columns = m/proc_columns;

    if( my_column < m%proc_columns)
    {
          num_columns++;
         j_start = num_columns * my_column + 1;
    }
    else
         j_start  = num_columns*my_column + n%proc_columns +1;

//    printf("%d:Local number of   rows and columns: %d,%d\n", me, num_rows, num_columns);
//    printf("%d:Local start index rows and columns: %d,%d\n", me, i_start, j_start);

    gr->lrow = num_rows;
    gr->lcol = num_columns;
    gr->start_row = i_start;
    gr->start_col = j_start;

   // allocate temporary arrays for exchange of columns

   right_col  = (float*) malloc((gr->lrow+2)*sizeof(float));
    left_col  = (float*) malloc((gr->lrow+2)*sizeof(float));

  }



}

// boundary exchange

void exchange(para* p, grid* gr)
{

  int i,j;

  MPI_Status status_100, status_200, status_300, status_400;

  // send down first row

  MPI_Send(gr->x_new[1], gr->lcol+2, MPI_FLOAT,gr->down,100,MPI_COMM_WORLD);

  MPI_Recv(gr->x_new[gr->lrow+1], gr->lcol+2, MPI_FLOAT, gr->up,100,MPI_COMM_WORLD,&status_100);

  // send up last row

  MPI_Send(gr->x_new[gr->lrow], gr->lcol+2, MPI_FLOAT,gr->up,200,MPI_COMM_WORLD);

  MPI_Recv(gr->x_new[0], gr->lcol+2, MPI_FLOAT, gr->down,200,MPI_COMM_WORLD,&status_200);

 // copy left column to tmp arrays

 if(gr->left != MPI_PROC_NULL)
 {
   for(i=0; i< gr->lrow+2; i++)
   {
     left_col[i] = gr->x_new[i][1];
   }

   MPI_Send(left_col, gr->lrow+2, MPI_FLOAT,gr->left,300,MPI_COMM_WORLD);
  }

  if(gr->right != MPI_PROC_NULL)
  {

   MPI_Recv(right_col, gr->lrow+2, MPI_FLOAT, gr->right,300,MPI_COMM_WORLD,&status_300);

  // copy right column to tmp
  // copy received left column to ghost cells


  for(i=0; i< gr->lrow+2; i++)
  {
    gr->x_new[i][gr->lcol+1] = right_col[i];
    right_col[i] = gr->x_new[i][gr->lcol];
  }

  // send right

  MPI_Send(right_col, gr->lrow+2, MPI_FLOAT,gr->right,400,MPI_COMM_WORLD);

  }

  if(gr->left != MPI_PROC_NULL)
  {

    MPI_Recv(left_col , gr->lrow+2, MPI_FLOAT, gr->left,400,MPI_COMM_WORLD,&status_400);

    for(i=0; i< gr->lrow+2; i++)
    {
      gr->x_new[i][0] = left_col[i];
    }

  }
}

// test program

/*

int main(int argc, char** argv)
{

 para p;
 grid gr;

 p.nprocs = atoi(argv[1]);
 p.me     = atoi(argv[2]);

 p.n = 1000;
 p.m = 1000;

 process_grid(&p,&gr);

}
*/

#endif
