/*

The example code {in this document / on this webpage} (hereafter Example 
Code) is subject to all of the following terms and conditions:

1. Intel Corporation ("Intel") grants to you a non-exclusive, non-
assignable copyright license to make only the minimum number of copies of 
the Example Code reasonably necessary for your internal use, and to 
modify the Example Code that are provided in source code (human readable) 
form.

2. You may not reverse-assemble, reverse-compile, or otherwise reverse-
engineer any software provided solely in binary form.
 
3. You may not distribute to any third party any portion of the Example 
Code in any form.
 
4. Title to the Example Code and all copies thereof remain with Intel or 
its suppliers.  The Example Code are copyrighted and are protected by 
United States copyright laws and international treaty provisions.  You 
will not remove any copyright notice from the Example Code.  You agree to 
prevent any unauthorized copying of the Example Code.  Except as 
expressly provided herein, Intel does not grant any express or implied 
right to you under Intel patents, copyrights, trademarks, or trade secret 
information.  Subject to Intels ownership of the Example Code, all 
right, title and interest in and to your modifications shall belong to 
you.
 
5. The Example Code are provided "AS IS" without warranty of any kind.  
INTEL OFFERS NO WARRANTY EITHER EXPRESS OR IMPLIED INCLUDING THOSE OF 
MERCHANTABILITY, NONINFRINGEMENT OF THIRD- PARTY INTELLECTUAL PROPERTY OR 
FITNESS FOR A PARTICULAR PURPOSE. NEITHER INTEL NOR ITS SUPPLIERS SHALL 
BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING, WITHOUT LIMITATION, 
DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF 
BUSINESS INFORMATION, OR OTHER LOSS) ARISING OUT OF THE USE OF OR 
INABILITY TO USE THE EXAMPLE CODE, EVEN IF INTEL HAS BEEN ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGES.  BECAUSE SOME JURISDICTIONS PROHIBIT THE 
EXCLUSION OR LIMITATION OF LIABILITY FOR CONSEQUENTIAL OR INCIDENTAL 
DAMAGES, THE ABOVE LIMITATION MAY NOT APPLY TO YOU.
 
6. THE EXAMPLE CODE ARE NOT DESIGNED, INTENDED, OR AUTHORIZED FOR USE IN 
ANY TYPE OF SYSTEM OR APPLICATION IN WHICH THE FAILURE OF THE EXAMPLE 
CODE COULD CREATE A SITUATION WHERE PERSONAL INJURY OR DEATH MAY OCCUR 
(E.G MEDICAL SYSTEMS, LIFE SUSTAINING OR LIFE SAVING SYSTEMS).  Should 
you use the Example Code for any such unintended or unauthorized use, you 
shall indemnify and hold Intel and its officers, subsidiaries and 
affiliates harmless against all claims, costs, damages, and expenses, and 
reasonable attorney fees arising out of, directly or indirectly, any 
claim of product liability, personal injury or death associated with such 
unintended or unauthorized use, even if such claim alleges that Intel was 
negligent regarding the design or manufacture of the part.
 
7. You agree that any material, information or other communication you 
transmit or post to an Intel website or provide to Intel under this 
Agreement will be considered non-confidential and non-proprietary 
("Communications").  Intel will have no obligations with respect to the 
Communications.  You agree that Intel and its designees will be free to 
copy, modify, create derivative works, publicly display, disclose, 
distribute, license and sublicense through multiple tiers of distribution 
and licensees, incorporate and otherwise use the Communications and all 
data, images, sounds, text, and other things embodied therein, including 
derivative works thereto, for any and all commercial or non-commercial 
purposes. You are prohibited from posting or transmitting to or from an 
Intel website or provide to Intel any unlawful, threatening, libelous, 
defamatory, obscene, pornographic, or other material that would violate 
any law.  If you wish to provide Intel with your confidential 
information, Intel requires a non-disclosure agreement (NDA) to receive 
such confidential information, so please contact your Intel 
representative to ensure the proper NDA is in place.
 
8. The term of this Agreement will commence on the date this Agreement is 
accepted by you and will continue until terminated.  You may terminate 
this Agreement at any time.  Intel may terminate this Agreement at any 
time with written notice to you should you breach any provision in this 
Agreement.  Upon termination, you will immediately destroy the Example 
Code along with any copies you have made.
 
9. U.S. GOVERNMENT RESTRICTED RIGHTS: The Example Code are provided with 
"RESTRICTED RIGHTS".  Use, duplication or disclosure by the Government is 
subject to restrictions set forth in FAR52.227-14 and DFAR252.227-7013 et 
seq. or its successor.  Use of the Materials by the Government 
constitutes acknowledgment of Intel's rights in them.
 
10. APPLICABLE LAWS: Any claim arising under or relating to this 
Agreement shall be governed by the internal substantive laws of the State 
of Delaware or federal courts located in Delaware, without regard to 
principles of conflict of laws.  You may not export the Example Code in 
violation of applicable export laws.

* Other names and brands may be claimed as the property of others

*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "parameter.h"
#include "grid.h"

// Interfaces

extern int parameter(int, char**, para*);
extern int print_parameter( para*);
extern void clean_up(para*,grid*);

extern void timer_start();
extern double timer_elapsed();

extern void setup_grid(para*, grid*);

extern void   copy(para*,grid*);
extern void   red(para*, grid*);
extern void black(para*, grid*);
extern float residuum(para*,grid*);

extern void print_result(para*, grid*, int, float);
extern void compare_result(para*, grid*, int, float);

extern void exchange(para*,grid*);


int main(int argc, char* argv[])
{

  int i,j,it;

  int n,m;
  int maxiter;
  int do_print;
  int do_compare;
  float eps;

  para p;
  grid gr;

  float resid = 0.0;

  double t1;

  int me=0;

/*
 *  MPI parameters
 *
 */

#ifdef USE_MPI
  MPI_Init(&argc,&argv);
#endif

  parameter(argc, argv, &p);

  me=p.me;

//  printf("my rank is: %d of %d procs\n", me,p.nprocs);

  if(me == 0 ) print_parameter(&p);

  maxiter    = p.maxiter;
  do_print   = p.do_print;
  do_compare = p.do_compare;
  eps        = p.eps;

  setup_grid(&p, &gr);

// Iterations

timer_start();

for(it=0; it < maxiter; it++)
{

  // copy last iteration state to backup array x_cur

     copy(&p,&gr);

#ifdef USE_MPI
     // boundary exchange

     exchange(&p,&gr);
#endif


  // RED update

     red(&p,&gr);

#ifdef USE_MPI
     // boundary exchange

     exchange(&p,&gr);
#endif

  // BLACK update

     black(&p, &gr);

#ifdef USE_MPI
     // boundary exchange

     exchange(&p,&gr);
#endif

  // residuum

   resid = residuum(&p, &gr);

  // if(me == 0 && it%100 == 0) printf("Residuum after iteration %d is %lf\n", it,resid);

   if(resid < eps ) break;

} // iteration it

t1 = timer_elapsed();

if(me == 0)
{
  printf("\nResiduum after %d iterations = %lf \n", it,resid);
  printf("Compute time = %lf [sec]\n", t1);
  printf("Perf         = %lf [GFlops]\n", (double) (4+4+2)*p.n*p.n*it/t1*1.e-9);
}

// print out result array

if(me == 0 && do_print)   print_result(&p,&gr,it,resid);

// compare with reference

if(me == 0 && do_compare) compare_result(&p,&gr,it,resid);

clean_up(&p,&gr);

#ifdef USE_MPI
MPI_Finalize();
#endif

}

// clean up

void clean_up(para* p, grid* g)
{
// deallocate memory etc.

 free(g->x_cur[0]);
 free(g->x_new[0]);

 free(g->x_cur);
 free(g->x_new);
}


// functions

//float f(float x,float y)
//{
//  return 0.0;
//}

float g(float x, float y)
{
  return x*x - y*y;
}
