/******************************************************************************
** Copyright 2013-2014 Intel Corporation All Rights Reserved.                **
**                                                                           **
** The source code, information and material ("Material") contained herein   **
** is owned by Intel Corporation or its suppliers or licensors, and title to **
** such Material remains with Intel Corporation or its suppliers or          **
** licensors.                                                                **
** The Material contains proprietary information of Intel or its suppliers   **
** and licensors. The Material is protected by worldwide copyright laws and  **
** treaty provisions. No part of the Material may be used, copied,           **
** reproduced, modified, published, uploaded, posted, transmitted,           **
** distributed or disclosed in any way without Intel's prior express written **
** permission. No license under any patent, copyright or other intellectual  **
** property rights in the material is granted to or conferred upon you,      **
** either expressly, by implication, inducement, estoppel or otherwise. Any  **
** license under such intellectual property rights must be express and       **
** approved by Intel in writing.                                             **
******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <mathimf.h>
#include <omp.h>


double t0=0., t1=0.;

#define TYPE float
#define SQRT sqrtf
#define PRECISION 1

#define VLENGTH 16
#define VLENGTH_BITS 4

TYPE   epsilon_sqr = 0.01;
double native_time;
TYPE   *Mass;
TYPE   *Position0, *Velocity0, *Acceleration0;
TYPE   *Position1, *Velocity1, *Acceleration1;
TYPE   *Position0_X, *Position0_Y, *Position0_Z;
TYPE   *Position1_X, *Position1_Y, *Position1_Z;
int    global_number_of_bodies;
int    global_number_of_bodies_per_thread;
int    nthreads;


void Initialize(void)
{
    int size_per_mass_array = 1*(global_number_of_bodies)*sizeof(TYPE);
    int i;

    Mass          = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position0_X   = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position0_Y   = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position0_Z   = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position1_X   = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position1_Y   = (float *) _mm_malloc(size_per_mass_array, 4096);
    Position1_Z   = (float *) _mm_malloc(size_per_mass_array, 4096);

    Position0     = (float *) _mm_malloc(3*size_per_mass_array, 4096);
    Position1     = (float *) _mm_malloc(3*size_per_mass_array, 4096);

    Velocity0     = (float *) _mm_malloc(3*size_per_mass_array, 4096);
    Velocity1     = (float *) _mm_malloc(3*size_per_mass_array, 4096);

    Acceleration0 = (float *) _mm_malloc(3*size_per_mass_array, 4096);
    Acceleration1 = (float *) _mm_malloc(3*size_per_mass_array, 4096);

    for(i=0; i<global_number_of_bodies; i++) Mass[i] = i%4;
    for(i=0; i<3*global_number_of_bodies; i++) Position0[i] = i%5;
    for(i=0; i<3*global_number_of_bodies; i++) Position1[i] = i%6;
    for(i=0; i<3*global_number_of_bodies; i++) Velocity0[i] = i%7;
    for(i=0; i<3*global_number_of_bodies; i++) Velocity1[i] = i%8;
    for(i=0; i<3*global_number_of_bodies; i++) Acceleration0[i] = 0; //i%9;
    for(i=0; i<3*global_number_of_bodies; i++) Acceleration1[i] = 0; //i%10;

    for(i=0; i<global_number_of_bodies; i++) Position0_X[i] = Position0[3*i+0];
    for(i=0; i<global_number_of_bodies; i++) Position0_Y[i] = Position0[3*i+1];
    for(i=0; i<global_number_of_bodies; i++) Position0_Z[i] = Position0[3*i+2];

    for(i=0; i<global_number_of_bodies; i++) Position1_X[i] = Position1[3*i+0];
    for(i=0; i<global_number_of_bodies; i++) Position1_Y[i] = Position1[3*i+1];
    for(i=0; i<global_number_of_bodies; i++) Position1_Z[i] = Position1[3*i+2];

}

void Checking(TYPE *lMass, TYPE *lPosition0, TYPE *lPosition1, TYPE *lVelocity0, TYPE *lVelocity1, TYPE *lAcceleration0, TYPE *lAcceleration1, TYPE *lPosition0_X, TYPE *lPosition0_Y, TYPE *lPosition0_Z, TYPE *lPosition1_X, TYPE *lPosition1_Y, TYPE *lPosition1_Z)
{
    int i;

    double sum = 0;

    for(i=0; i<3*global_number_of_bodies; i++) sum += lAcceleration1[i];

    printf("Debug Sum = %lf\n", sum);
}

#pragma omp declare target
void Execute(int starting_index, int ending_index, int time_instance, int threadid)
{
    TYPE *Input_Position;
    TYPE *Output_Position;
    TYPE *Input_Velocity;
    TYPE *Output_Velocity;
    TYPE *Input_Acceleration;
    TYPE *Output_Acceleration;
    TYPE *Input_Position_X;
    TYPE *Input_Position_Y;
    TYPE *Input_Position_Z;

    if (!(time_instance%2)) {
        Input_Position = Position0;
        Input_Velocity = Velocity0;
        Input_Acceleration = Acceleration0;
        Output_Position = Position1;
        Output_Velocity = Velocity1;
        Output_Acceleration = Acceleration1;
        Input_Position_X = Position0_X;
        Input_Position_Y = Position0_Y;
        Input_Position_Z = Position0_Z;
    }
    else {
        Input_Position = Position1;
        Input_Velocity = Velocity1;
        Input_Acceleration = Acceleration1;
        Output_Position = Position0;
        Output_Velocity = Velocity0;
        Output_Acceleration = Acceleration0;
        Input_Position_X = Position1_X;
        Input_Position_Y = Position1_Y;
        Input_Position_Z = Position1_Z;
    }

/////////////////// NBody main loop start /////////////////////

#include "nbody_loop.cpp"

/////////////////// NBody main loop end /////////////////////
}

#pragma omp declare target
void * Compute_Acceleration(void *arg1)
{
    int threadid = (int)((size_t)(arg1));

    int starting_index = 0;
    int ending_index = global_number_of_bodies;

    Execute(starting_index, ending_index, 4, threadid);

    return NULL;
}

void Perform_NBody(void)
{

#  pragma omp target device(0)
   {
    int i;
    double t2=0., t3=0.;

    t2 = omp_get_wtime();

    Compute_Acceleration(0);

    t3 = omp_get_wtime();
    native_time = t3-t2;
   }

}

int main(int argc, char **argv)
{
    int i;

    if (argc != 3)
    {
        printf("Usage: nbody <number_of_bodies> <nthreads>\n");
        exit(123);
    }

    sscanf(argv[1], "%d", &global_number_of_bodies);
    sscanf(argv[2], "%d", &nthreads);

    printf("Number of Bodies = %d\n", global_number_of_bodies);

    Initialize();

    omp_set_num_threads(nthreads);

    Perform_NBody(); // Warm icache and dcache

    // Reinitialize the output arrays after the warmup run to get
    // the right debug sum in the end
    for(i=0; i<3*global_number_of_bodies; i++) Acceleration0[i] = 0; //i%9;
    for(i=0; i<3*global_number_of_bodies; i++) Acceleration1[i] = 0; //i%10;

    t0 = omp_get_wtime();

    Perform_NBody();

    t1 = omp_get_wtime();

    double clock_time_taken = t1-t0;

    Checking(Mass, Position0, Position1, Velocity0, Velocity1, Acceleration0, Acceleration1, Position0_X, Position0_Y, Position0_Z, Position1_X, Position1_Y, Position1_Z);

    printf("\nRun time = %f sec. "
    "computation time = %f sec., "
    "offload overhead = %f sec.\n",
    clock_time_taken,
    native_time,
    clock_time_taken - native_time);
}
