/******************************************************************************
** Copyright 2013-2014 Intel Corporation All Rights Reserved.                **
**                                                                           **
** The source code, information and material ("Material") contained herein   **
** is owned by Intel Corporation or its suppliers or licensors, and title to **
** such Material remains with Intel Corporation or its suppliers or          **
** licensors.                                                                **
** The Material contains proprietary information of Intel or its suppliers   **
** and licensors. The Material is protected by worldwide copyright laws and  **
** treaty provisions. No part of the Material may be used, copied,           **
** reproduced, modified, published, uploaded, posted, transmitted,           **
** distributed or disclosed in any way without Intel's prior express written **
** permission. No license under any patent, copyright or other intellectual  **
** property rights in the material is granted to or conferred upon you,      **
** either expressly, by implication, inducement, estoppel or otherwise. Any  **
** license under such intellectual property rights must be express and       **
** approved by Intel in writing.                                             **
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <omp.h>

int ImageHeight;
int ImageWidth;
int max_iter;
double t0, t1;
float *vals;
float _Complex *cvals;
unsigned int *res;

void Mandelbrot(void);

int main()
{
  ImageHeight = 1024 * 4;
  ImageWidth = 1024 * 4;
  max_iter = 30;
  int i, j;
  srand(1234);
  
  vals = (float*)_mm_malloc(ImageHeight*ImageWidth * sizeof(float), 64);
  cvals = (float _Complex*)_mm_malloc(ImageHeight*ImageWidth * sizeof(float _Complex), 64);
  res = (unsigned int*)_mm_malloc(ImageHeight*ImageWidth * sizeof(unsigned int), 64);
  fprintf(stderr, "mandelbrot problem size: %d x %d\n", ImageHeight, ImageWidth);
  
  for (i = 0; i < ImageHeight; i++) {
    for (j = 0; j < ImageWidth; j++) {
      vals[i*ImageWidth+j] = rand() / (float)RAND_MAX;
    }
  }
  for (i = 0; i < ImageHeight; i++) {
    for (j = 0; j < ImageWidth; j++) {
      cvals[i*ImageWidth+j] = (vals[i] + _Complex_I * vals[j]);
    }
  }
  
  //warm up
  Mandelbrot();

  t0 = omp_get_wtime();
  for (i=0; i<5; i++){
    Mandelbrot();
  }
  t1 = omp_get_wtime();
  printf("\nRun time = %f sec.\n", t1 - t0);
  
  _mm_free(vals);
  _mm_free(cvals);
  _mm_free(res);
  return 0;
}
