
/******************************************************************************
** Copyright 2013 Intel Corporation All Rights Reserved.                     **
**                                                                           **
** The source code, information and material ("Material") contained herein   **
** is owned by Intel Corporation or its suppliers or licensors, and title to **
** such Material remains with Intel Corporation or its suppliers or          **
** licensors.                                                                **
** The Material contains proprietary information of Intel or its suppliers   **
** and licensors. The Material is protected by worldwide copyright laws and  **
** treaty provisions. No part of the Material may be used, copied,           **
** reproduced, modified, published, uploaded, posted, transmitted,           **
** distributed or disclosed in any way without Intel's prior express written **
** permission. No license under any patent, copyright or other intellectual  **
** property rights in the material is granted to or conferred upon you,      **
** either expressly, by implication, inducement, estoppel or otherwise. Any  **
** license under such intellectual property rights must be express and       **
** approved by Intel in writing.                                             **
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>
#include <math.h>

extern int ImageHeight;
extern int ImageWidth;
extern int max_iter;
extern unsigned int *res;
extern float _Complex *cvals;

// Compute the number of iterations required before the Mandlebrot
// iterations from the given starting value cause divergence (escape 
// from a circle of radius 2), or until max_iters is reached.   
unsigned int mandel(float _Complex c)
{
  unsigned int count = 1;
  float _Complex z = c;
  int j;
  for (j = 0; j < max_iter; j++) {
    z = z * z + c;
    float absZ = cabsf(z);
    if (absZ >= 2.0f) {
      break;
    }
    count++;
  }

  return count;
}

// Compute the Mandlebrot escape counts for all samples in a region of the 
// complex plane specified in the "scale" array.
void Mandelbrot(void)
{
  int x, y;

  for (y = 0; y < ImageHeight; y++) {
    for (x = 0; x < ImageWidth; x++) {
      res[y*ImageWidth + x] = mandel(cvals[y*ImageWidth + x]);
    }
  }
}
